package xyz.lychee.lagfixer.libs.adventure.platform.facet;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import xyz.lychee.lagfixer.libs.adventure.audience.Audience;
import xyz.lychee.lagfixer.libs.adventure.audience.ForwardingAudience;
import xyz.lychee.lagfixer.libs.adventure.identity.Identity;
import xyz.lychee.lagfixer.libs.adventure.key.Key;
import xyz.lychee.lagfixer.libs.adventure.permission.PermissionChecker;
import xyz.lychee.lagfixer.libs.adventure.platform.AudienceProvider;
import xyz.lychee.lagfixer.libs.adventure.pointer.Pointered;
import xyz.lychee.lagfixer.libs.adventure.pointer.Pointers;
import xyz.lychee.lagfixer.libs.adventure.text.renderer.ComponentRenderer;
import xyz.lychee.lagfixer.libs.adventure.util.TriState;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.ApiStatus.Internal;

@Internal
public abstract class FacetAudienceProvider<V, A extends FacetAudience<V>> implements AudienceProvider, ForwardingAudience {
   protected static final Locale DEFAULT_LOCALE;
   protected final ComponentRenderer<Pointered> componentRenderer;
   private final Audience console;
   private final Audience player;
   protected final Map<V, A> viewers;
   private final Map<UUID, A> players;
   private final Set<A> consoles;
   private A empty;
   private volatile boolean closed;

   protected FacetAudienceProvider(final @NotNull ComponentRenderer<Pointered> componentRenderer) {
      this.componentRenderer = (ComponentRenderer)Objects.requireNonNull(componentRenderer, "component renderer");
      this.viewers = new ConcurrentHashMap();
      this.players = new ConcurrentHashMap();
      this.consoles = new CopyOnWriteArraySet();
      this.console = new ForwardingAudience() {
         public @NotNull Iterable<? extends Audience> audiences() {
            return FacetAudienceProvider.this.consoles;
         }

         public @NotNull Pointers pointers() {
            return FacetAudienceProvider.this.consoles.size() == 1 ? ((FacetAudience)FacetAudienceProvider.this.consoles.iterator().next()).pointers() : Pointers.empty();
         }
      };
      this.player = Audience.audience(this.players.values());
      this.closed = false;
   }

   public void addViewer(@NotNull final V viewer) {
      if (!this.closed) {
         A audience = (FacetAudience)this.viewers.computeIfAbsent(Objects.requireNonNull(viewer, "viewer"), (v) -> {
            return this.createAudience(Collections.singletonList(v));
         });
         FacetPointers.Type type = (FacetPointers.Type)audience.getOrDefault(FacetPointers.TYPE, FacetPointers.Type.OTHER);
         if (type == FacetPointers.Type.PLAYER) {
            UUID id = (UUID)audience.getOrDefault(Identity.UUID, (Object)null);
            if (id != null) {
               this.players.putIfAbsent(id, audience);
            }
         } else if (type == FacetPointers.Type.CONSOLE) {
            this.consoles.add(audience);
         }

      }
   }

   public void removeViewer(@NotNull final V viewer) {
      A audience = (FacetAudience)this.viewers.remove(viewer);
      if (audience != null) {
         FacetPointers.Type type = (FacetPointers.Type)audience.getOrDefault(FacetPointers.TYPE, FacetPointers.Type.OTHER);
         if (type == FacetPointers.Type.PLAYER) {
            UUID id = (UUID)audience.getOrDefault(Identity.UUID, (Object)null);
            if (id != null) {
               this.players.remove(id);
            }
         } else if (type == FacetPointers.Type.CONSOLE) {
            this.consoles.remove(audience);
         }

         audience.close();
      }
   }

   public void refreshViewer(final @NotNull V viewer) {
      A audience = (FacetAudience)this.viewers.get(viewer);
      if (audience != null) {
         audience.refresh();
      }

   }

   protected abstract @NotNull A createAudience(final @NotNull Collection<V> viewers);

   public @NotNull Iterable<? extends Audience> audiences() {
      return this.viewers.values();
   }

   public @NotNull Audience all() {
      return this;
   }

   public @NotNull Audience console() {
      return this.console;
   }

   public @NotNull Audience players() {
      return this.player;
   }

   public @NotNull Audience player(final @NotNull UUID playerId) {
      return (Audience)this.players.getOrDefault(playerId, this.empty());
   }

   private @NotNull A empty() {
      if (this.empty == null) {
         this.empty = this.createAudience(Collections.emptyList());
      }

      return this.empty;
   }

   public @NotNull Audience filter(final @NotNull Predicate<V> predicate) {
      return Audience.audience(filter(this.viewers.entrySet(), (entry) -> {
         return predicate.test(entry.getKey());
      }, Map.Entry::getValue));
   }

   private @NotNull Audience filterPointers(final @NotNull Predicate<Pointered> predicate) {
      return Audience.audience(filter(this.viewers.entrySet(), (entry) -> {
         return predicate.test((Pointered)entry.getValue());
      }, Map.Entry::getValue));
   }

   public @NotNull Audience permission(final @NotNull String permission) {
      return this.filterPointers((pointers) -> {
         return ((PermissionChecker)pointers.get(PermissionChecker.POINTER).orElse(PermissionChecker.always(TriState.FALSE))).test(permission);
      });
   }

   public @NotNull Audience world(final @NotNull Key world) {
      return this.filterPointers((pointers) -> {
         return world.equals(pointers.getOrDefault(FacetPointers.WORLD, (Object)null));
      });
   }

   public @NotNull Audience server(final @NotNull String serverName) {
      return this.filterPointers((pointers) -> {
         return serverName.equals(pointers.getOrDefault(FacetPointers.SERVER, (Object)null));
      });
   }

   public void close() {
      this.closed = true;
      Iterator var1 = this.viewers.keySet().iterator();

      while(var1.hasNext()) {
         V viewer = var1.next();
         this.removeViewer(viewer);
      }

   }

   private static <T, V> @NotNull Iterable<V> filter(final @NotNull Iterable<T> input, final @NotNull Predicate<T> filter, final @NotNull Function<T, V> transformer) {
      return new Iterable<V>() {
         public @NotNull Iterator<V> iterator() {
            return new Iterator<V>() {
               private final Iterator<T> parent = input.iterator();
               private V next;

               {
                  this.populate();
               }

               private void populate() {
                  this.next = null;

                  Object next;
                  do {
                     if (!this.parent.hasNext()) {
                        return;
                     }

                     next = this.parent.next();
                  } while(!filter.test(next));

                  this.next = transformer.apply(next);
               }

               public boolean hasNext() {
                  return this.next != null;
               }

               public V next() {
                  if (this.next == null) {
                     throw new NoSuchElementException();
                  } else {
                     V next = this.next;
                     this.populate();
                     return next;
                  }
               }
            };
         }

         public void forEach(final Consumer<? super V> action) {
            Iterator var2 = input.iterator();

            while(var2.hasNext()) {
               T each = var2.next();
               if (filter.test(each)) {
                  action.accept(transformer.apply(each));
               }
            }

         }
      };
   }

   static {
      DEFAULT_LOCALE = Locale.US;
   }
}
